/*
 * $Header: /home/vikas/src/xtacacsd/RCS/misc.c,v 1.4 1996/05/19 17:26:24 vikas Exp $
 */

/*
 * Miscellanous routines for xtacacsd.
 *
 * uwtmp_entry():
 *	Updates the utmp and wtmp files.
 *
 * skip_spaces():
 *	Given a string pointer, returns a pointer after skipping the
 * 	leading spaces.
 *
 * Strdup():
 *	Duplicate a string (like strdup).
 *
 * report():
 *	Like syslog(), but prints to stderr, looks at logging, etc.
 *
 *
 *	-Vikas Aggarwal, vikas@navya.com, June 1995
 *
 * $Log: misc.c,v $
 * Revision 1.4  1996/05/19  17:26:24  vikas
 * Added routines to check the size of the utmp/wtmp files and
 * match with the currently compiled utmp structure while
 * updating the files.
 *
 * Revision 1.3  1996/04/07  19:03:09  vikas
 * Added support for LOGOUT_CHAR in the username instead of having
 * only a null string in the username for logouts.
 * Added macros to make things look nicer. Also checks to see
 * the size of the file when opening and warns if wrong size.
 *
 * Revision 1.2  1995/11/10  22:10:03  vikas
 * Fixed lseek() bug for BSDI machines ; subtle but created large
 * utmp files.
 *
 * Revision 1.1  1995/07/14  16:18:47  vikas
 * Initial revision
 *
 *
 */


#ifndef lint
 static char rcsid[] = "$RCSfile: misc.c,v $ $Revision: 1.4 $ $Date: 1996/05/19 17:26:24 $" ;
 static char sccsid[] = "$RCSfile: misc.c,v $ $Revision: 1.4 $ $Date: 1996/05/19 17:26:24 $" ;
#endif

#include <sys/types.h>
#include <stdio.h>
#ifdef __STDC__
# include <stdarg.h>		/* ANSI C, variable length args for report() */
#else
# include <varargs.h>		/* has 'vararg' definitions */
#endif
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <syslog.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>		/* for fstat */
#include <netinet/in.h>		/* htonl() definitions */

#include "tacutmp.h"
#include "tacacs.h"

#define EXTERN extern
#include "common.h"
#undef EXTERN

/*+ report:
 *
 * This routine reports errors and such via stderr and syslog() if
 * appopriate.  It just helps avoid a lot of if-else in the code.
 *
 * The syntax is identical to syslog(3), but %m is not considered special
 * for output to stderr (i.e. you'll see "%m" in the output. . .).  Also,
 * control strings should normally end with \n since newlines aren't
 * automatically generated for stderr output (whereas syslog strips out all
 * newlines and adds its own at the end).
 *
 * Print to stderr only in debug+standalone (otherwise the process closes
 * the stderr, etc. file descriptors on startup).
 */

#ifdef __STDC__
report (int priority, char *fmt, ...)
#else
report(priority, fmt, va_alist)
     int priority;
     char *fmt;
     va_dcl			/* no terminating semi-colon */
#endif
{
    char msg[256];		/* temporary string */
    va_list ap;

#ifdef __STDC__
    va_start(ap, fmt);
#else
    va_start(ap);
#endif
    vsprintf (msg, fmt, ap);
    va_end(ap);

    if (stand)
      fprintf (stderr, "(xtacacsd) ") ;

    /*
     * if compiled with DEBUG, log DEBUG messages at the INFO level
     */
       
    if (priority == LOG_DEBUG)
    {
#ifdef DEBUG
	if (debug && logging)
	  syslog(LOG_INFO, "%s", msg);
#endif
	if (debug && stand)
	  fprintf(stderr, "%s\n", msg);

    }

    else if (priority == LOG_INFO)	/* log only if logging */
    {
	if (logging)
	  syslog(LOG_INFO, "%s", msg);
	if (debug && stand)
	  fprintf(stderr,  "%s\n", msg);

    }
    else		/* more urgent */
    {
	syslog(priority, "%s", msg);
	if ( (priority == LOG_NOTICE && stand) || priority < LOG_NOTICE)
	  fprintf(stderr, "%s\n", msg);
    }

    return ;
}	/* end report() */



/*
 * Function to test the size of the supplied file and see if the tacutmp
 * record size is evenly divisible. Else, try and assume an old style file
 * and delete the size of the UT_COMMENT and try again.
 * Required for backward compatibility.
 */
static int 
get_utmp_entrysize(fd)
     int fd;		/* opened file descriptor */
{
    struct stat stb;
    struct tacutmp entry;
    int entrysize;

    entrysize = sizeof entry;

    if (fd < 0)
      return (-1);

    if (fstat(fd, &stb) == -1) {
	report(LOG_ERR, "Error fstat-ing file %s\n", sys_errlist[errno]);
	return (-1);
    }

    if ( (stb.st_size % entrysize) != 0)
    {
#ifdef XTACUTMP		/* new tacupd, old format utmp file */
	entrysize = entrysize - UT_COMMENTSIZE ;
	if (stb.st_size % entrysize )
	  return (-1);	/* should be treated as fatal */
#else	/* old tacupd, new format utmp file, so just exit */
	return (-1);
#endif	/* XTACUTMP */

    }	/* if (stb.st_size(...) */

    return (entrysize);
}



/*+ uwtmp_entry:
 *
 * Updates the 'wtmp_file' and utmp files.
 *
 * Called by 'slipOn' and 'slipOff' with line types of SLIP_TTY.
 * If the 'name' is a null string or begins with a UT_LOGOUT_CHAR
 * (in logout's), the corresponding entry in the utmp file is flushed.
 *
 * The calling program MUST initialize the wtmpfd & utmpfd file descriptors
 * to -1 so that this function attempts to open them.
 */

#define lineEq(A, B)	(!strncmp(A.ut_line, B.ut_line, UT_LINESIZE))
#define nameEq(A, B)	(!strncmp(A.ut_name, B.ut_name, UT_NAMESIZE))
#define hostEq(A, B)	(!strncmp(A.ut_host, B.ut_host, UT_HOSTSIZE))

uwtmp_entry (line, name, host, timestamp, comment)
     char *line ;		/* Incoming line for user */
     char *name;		/* User's login name */
     char *host;		/* name/ip of host sending this info */
     unsigned long timestamp;	/* login time, can be zero */
     char *comment;		/* for extended xtacutmp structure */
{
    struct tacutmp oentry, entry;
    register char *c ;
    static int  urecsize, wrecsize;	/* current size of the utmp entries */
    static int  dofsync = 0;		/* dont FSYNC every time */
    static int retryopen = -1;

    interruptible = 0;			/* no alarms allowed here */

    if (utmpfd < 0  || wtmpfd < 0)
      retryopen = ++retryopen % 10 ;	/* retry opening every 10th time */

    if (wtmpfd < 0 && wtmp_file != NULL && retryopen == 0 )
      if ((wtmpfd = open(wtmp_file, O_WRONLY | O_CREAT | O_APPEND, 0664)) < 0)
	report(LOG_ERR, "Can't open wtmp file '%s': %s",
	       wtmp_file, sys_errlist[errno]);

    if (utmpfd < 0 && utmp_file != NULL && retryopen == 0)
      if ((utmpfd = open(utmp_file, O_RDWR | O_CREAT, 0664)) < 0)
	report(LOG_ERR, "Can't open utmp file '%s': %s",
	       utmp_file, sys_errlist[errno]);

    
    if (utmpfd < 0 && wtmpfd < 0) {
	interruptible = 1;
	return ;
    }

    if (wtmpfd >= 0 && wrecsize <= 0) 	/* get record size */
      if ( (wrecsize = get_utmp_entrysize(wtmpfd)) <= 0 ) {
	  report(LOG_ALERT,
		 "wtmp file format not compatible with compiled tacutmp size, CORRECT IMMEDIATELY USING XTACUTMP DEFINE, NOT WRITING WTMP\n");
	  close(wtmpfd);
	  wtmpfd = -1;
      }

    if (utmpfd >= 0 && urecsize <= 0)
      if ( (urecsize = get_utmp_entrysize(utmpfd)) <= 0 ) {
	  report(LOG_ALERT,
		 "utmp file format not compatible with compiled tacutmp size, CORRECT IMMEDIATELY USING XTACUTMP DEFINE, NOT WRITING UTMP\n");
	  close(utmpfd);
	  utmpfd = -1;
      }


    bzero(&entry, sizeof entry);

    /* Convert the hostname to all lowercase (DNS responses are case
     * sensitive even though the query is not.
     */
    for (c = host ; *c != '\0' ; ++c)
      *c = tolower(*c) ;

      /* notice no terminating null in the following if no space for it */
#define SCOPY(A,B) \
  { \
      if (strlen(A) < sizeof B) \
	strcpy(B, A); \
      else bcopy(A, B, sizeof B); \
  }
    SCOPY(line, entry.ut_line);
    SCOPY(name, entry.ut_name);
    SCOPY(host, entry.ut_host);

    if (timestamp) 
      entry.ut_time = timestamp;
    else
      entry.ut_time = time(NULL);	/* fill with current time if 0 */

#ifdef XTACUTMP
    SCOPY(comment, entry.ut_comment);
#endif
    
#ifdef DEBUG
    report(LOG_DEBUG, "uwtmp_entry: '%s'@%s %s, time= %ld, comment= '%s' ",
	   name, line, host, entry.ut_time, comment);
#endif

    if (wtmpfd >= 0)	/* file was opened in append mode, so just write out */
    {
      if (write(wtmpfd, &entry, wrecsize) != wrecsize)
	report(LOG_ERR, "uwtmp_entry: couldn't write wtmp file '%s', %s",
	       wtmp_file, sys_errlist[errno]);
#ifndef NOFSYNC
      else
	if ((dofsync = ++dofsync % 20) == 0)	/* sync every 20 times */
	  fsync (wtmpfd);
#endif
    }	/* endif wtmpfd >= 0 */
#ifdef DEBUG
    else
      report(LOG_DEBUG, "uwtmp_entry: wtmp entry not written, unopened fd");
#endif

    /*
     * Update alternate wtmp file (with the hostname tacked on) so
     * that 'last' etc. can handle it.
     */
    if (update_host_wtmp)
    {
	char hwtmpfile[BUFSIZ] ;	/* alternate file to update */
	int  hwtmpfd ;

	strcpy(hwtmpfile, wtmp_file);	/* same parent wtmp name */
	strcat(hwtmpfile, ".");		/* add a dot */

	/* Create a sensible filename  and add entry */

	if (*from_hname < '0' || *from_hname > '9')  /* Not an IP address,,, */
	  strncat(hwtmpfile, from_hname, strcspn(from_hname, ".")) ;   /* only header */
	else
	  strncat(hwtmpfile, from_hname, strlen(from_hname)) ;
    
	hwtmpfile[sizeof(hwtmpfile) -1] = '\0';		/* terminating NULL */
	
	hwtmpfd = open(hwtmpfile, O_WRONLY | O_CREAT | O_APPEND, 0664);
	if (hwtmpfd < 0) {
	    report(LOG_ERR, "uwtmp_entry: couldn't open host wtmp- %s",
		   hwtmpfile);
	}
	else {
	    if (write(hwtmpfd, &entry, wrecsize) != wrecsize)
	      report(LOG_ERR, "uwtmp_entry: write host wtmp %s- %s",
		     hwtmpfile, sys_errlist[errno]);
	    else
	      close(hwtmpfd) ;	/* close the alternate wtmp file */
	}

    }	/* end update_host_wtmp */

    
    /*
     * Now update the utmp file
     */
    if (utmpfd < 0)			/* utmp file not open */
    {
	interruptible = 1;
	return;
    }

    lseek (utmpfd, 0, SEEK_SET);		/* rewind file */

    /*
     * new utmp entry. Make a attempt to avoid duplicates. However,
     * cannot really avoid unless we search thru entire file each
     * time to make sure that the entry does not exist further up.
     */
    if (*name && *name != UT_LOGOUT_CHAR)
    {
	int n;

	while ((n = read(utmpfd, (char *)&oentry, urecsize)) == urecsize)
	{		/* write in erased entry */
	    if (*oentry.ut_name == '\0' || *oentry.ut_name == UT_LOGOUT_CHAR)
	    {
		/* back up one entry */
		lseek(utmpfd, -(off_t)urecsize, SEEK_CUR);
		if (write(utmpfd, (char *)&entry,urecsize) != urecsize)
		  report(LOG_ERR, "utmp_entry: couldn't write utmp file, %s",
			 sys_errlist[errno]);
#ifdef DEBUG
		else
		  report (LOG_DEBUG, "utmp_entry: writing in erased utmp entry");
#endif
		break;	/* out of while() */
	    }
	    else if (hostEq(oentry, entry) && lineEq(oentry, entry))
	    {
		report(LOG_INFO, "utmp_entry: overwriting existing entry for %.*s on %s@%.*s",
		       sizeof oentry.ut_name, oentry.ut_name, entry.ut_line,
		       sizeof entry.ut_host, entry.ut_host);
		/* back up one entry */
		lseek(utmpfd, -(off_t)urecsize, SEEK_CUR);
		if (write(utmpfd, (char *)&entry, urecsize) != urecsize)
		  report(LOG_ERR, "utmp_entry: couldn't write utmp file, %s",
			 sys_errlist[errno]);
	      break;
	    }
	}	/* end while() */


	if (n != urecsize)	/* end of file, so write new entry here */
	{
	    if (write(utmpfd, (char *)&entry, urecsize) != urecsize)
	      report(LOG_ERR, "utmp_entry: couldn't write utmp file, %s",
		     sys_errlist[errno]);
#ifdef DEBUG
	    else
	      report (LOG_DEBUG, "utmp_entry: writing utmp entry at end of file");
#endif
	}

    }	/* end: if (*name) */

    else	/* erasing an old entry, name null. Line is ~ on reload */
    {
	while (read(utmpfd, (char *)&oentry, urecsize) == urecsize)
	{
	    if (hostEq(oentry, entry))
	      if (*entry.ut_line == '~' || lineEq(oentry, entry))
	      {
		  /* back up one entry */
#ifdef DEBUG
		  report (LOG_DEBUG, "utmp_entry: deleting %.*s %.*s %.*s",
			  sizeof oentry.ut_host, oentry.ut_host,
			  sizeof oentry.ut_name, oentry.ut_name,
			  sizeof oentry.ut_line, oentry.ut_line);
#endif
		  lseek(utmpfd, -(off_t)urecsize, SEEK_CUR);
		  bzero(oentry.ut_line, sizeof oentry.ut_line);
		  bzero(oentry.ut_name, sizeof oentry.ut_name);	
		  bzero(oentry.ut_host, sizeof oentry.ut_host);

		  if (write(utmpfd, (char *)&oentry, urecsize) != urecsize)
		  {
		      report(LOG_ERR, "utmp_entry: write- %s", 
			     sys_errlist[errno]);
		      break;
		  }

	      }	/* end: if(same host + line) */

	}	/* end: while() */


    }	/* end: if-else new/erase entry */

#ifndef NOFSYNC
    fsync(utmpfd);
#endif

    interruptible = 1;

}	/* end: uwtmp_entry */

/*
 * Count how many times the given user is logged in.
 */
cur_login_count(user)
     char *user;
{
    int count = 0;
    static int recsize ;
    struct tacutmp entry ;	/* to read from the utmp file */

    if (!recsize)
      if ( (recsize = get_utmp_entrysize(utmpfd)) < 0 ) {
	  report(LOG_ALERT, "cur_login_count: get_utmp_entrysize() ret -1, FIX COMPILE TIME DEFINE XTACUTMP");
	  return (-1);
      }

    lseek(utmpfd, 0, SEEK_SET);	/* rewind */
    while (read(utmpfd, (char *)&entry, recsize) == recsize) 
    {
	if (strncmp(entry.ut_name, user, sizeof entry.ut_name))
	  continue;	/* no match */
	else
	  ++count;
    }
    return(count);
}



/*+ skip_spaces
 *
 * 	Skip leading spaces in the string passed to it and return back
 * a char pointer.
 */
char *skip_spaces(str)
     char *str ;
{
    if (str == NULL)
      return (str);
    while (*str == ' '  || *str == '\t')
      ++str ;                           /* skip leading spaces */
 
    return(str) ;
}

/*+ Strdup:
 *	Ultrix doesn't have a strdup(), so here's one...
 *
 */
char *Strdup(s)
     register char *s ;
{
    char *t ;
    
    if (s == NULL)
    {
        t = (char *)malloc(1);
        *t = '\0';
    }
    else
    {
        t = (char *)malloc(strlen(s) + 1);
        if (t != NULL)
          (char *)strcpy(t, s);
        else
	{
	    perror("Strdup: malloc()");
	    exit(1);
	}
    }
    
    return (t);
}	/* end: Strdup() */



